@@ -12,13 +12,13 @@ import subprocess
 import sys
 import time
 from asyncio.subprocess import Process
+from importlib.metadata import PackageNotFoundError, distribution
 from pathlib import Path
 from typing import List, Optional, Union
 from uuid import uuid4
 
 import aiohttp
 import requests
-from typing_extensions import Self
 from websockets.client import WebSocketClientProtocol
 from websockets.client import connect as ws_connect
 from websockets.exceptions import ConnectionClosedError
@@ -37,7 +37,7 @@ class LocalBox(BaseBox):
     This is useful for testing and development.
     """
 
-    _instance: Optional[Self] = None
+    _instance: Optional["LocalBox"] = None
 
     def __new__(cls, *args, **kwargs):
         if not cls._instance:
@@ -58,7 +58,7 @@ class LocalBox(BaseBox):
         self.kernel_id: Optional[dict] = None
         self.ws: Union[WebSocketClientProtocol, ClientConnection, None] = None
         self.jupyter: Union[Process, subprocess.Popen, None] = None
-        self.session: Optional[aiohttp.ClientSession] = None
+        self.aiohttp_session: Optional[aiohttp.ClientSession] = None
 
     def start(self) -> CodeBoxStatus:
         os.makedirs(".codebox", exist_ok=True)
@@ -68,6 +68,7 @@ class LocalBox(BaseBox):
             out = None
         else:
             out = subprocess.PIPE
+        self._check_installed()
         try:
             python = Path(sys.executable).absolute()
             self.jupyter = subprocess.Popen(
@@ -119,34 +120,57 @@ class LocalBox(BaseBox):
         except requests.exceptions.ConnectionError:
             pass
         else:
-            if response.status_code == 200:
-                self.port += 1
-                self._check_port()
+            try:
+                requests.post(f"http://localhost:{self.port}/api/shutdown")
+            except requests.exceptions.ConnectionError:
+                if response.status_code == 200:
+                    self.port += 1
+                    self._check_port()
+
+    def _check_installed(self) -> None:
+        try:
+            distribution("jupyter-kernel-gateway")
+        except PackageNotFoundError:
+            print(
+                "Make sure 'jupyter-kernel-gateway' is installed "
+                "when using without a CODEBOX_API_KEY.\n"
+                "You can install it with 'pip install jupyter-kernel-gateway'."
+            )
+            raise
 
     async def astart(self) -> CodeBoxStatus:
         os.makedirs(".codebox", exist_ok=True)
-        self.session = aiohttp.ClientSession()
+        self.aiohttp_session = aiohttp.ClientSession()
         await self._acheck_port()
         if settings.VERBOSE:
             print("Starting kernel...")
             out = None
         else:
             out = asyncio.subprocess.PIPE
+        self._check_installed()
         python = Path(sys.executable).absolute()
-        self.jupyter = await asyncio.create_subprocess_exec(
-            python,
-            "-m",
-            "jupyter",
-            "kernelgateway",
-            "--KernelGatewayApp.ip='0.0.0.0'",
-            f"--KernelGatewayApp.port={self.port}",
-            stdout=out,
-            stderr=out,
-            cwd=".codebox",
-        )
+        try:
+            self.jupyter = await asyncio.create_subprocess_exec(
+                python,
+                "-m",
+                "jupyter",
+                "kernelgateway",
+                "--KernelGatewayApp.ip='0.0.0.0'",
+                f"--KernelGatewayApp.port={self.port}",
+                stdout=out,
+                stderr=out,
+                cwd=".codebox",
+            )
+        except Exception as e:
+            print(e)
+            raise ModuleNotFoundError(
+                "Jupyter Kernel Gateway not found, please install it with:\n"
+                "`pip install jupyter_kernel_gateway`\n"
+                "to use the LocalBox."
+            )
         while True:
             try:
-                response = await self.session.get(self.kernel_url)
+                response = await self.aiohttp_session.get(self.kernel_url)
                 if response.status == 200:
                     break
             except aiohttp.ClientConnectorError:
@@ -157,7 +181,7 @@ class LocalBox(BaseBox):
                 print("Waiting for kernel to start...")
             await asyncio.sleep(1)
 
-        response = await self.session.post(
+        response = await self.aiohttp_session.post(
             f"{self.kernel_url}/kernels", headers={"Content-Type": "application/json"}
         )
         self.kernel_id = (await response.json())["id"]
@@ -169,9 +193,9 @@ class LocalBox(BaseBox):
 
     async def _acheck_port(self) -> None:
         try:
-            if self.session is None:
-                self.session = aiohttp.ClientSession()
-            response = await self.session.get(f"http://localhost:{self.port}")
+            if self.aiohttp_session is None:
+                self.aiohttp_session = aiohttp.ClientSession()
+            response = await self.aiohttp_session.get(f"http://localhost:{self.port}")
         except aiohttp.ClientConnectorError:
             pass
         except aiohttp.ServerDisconnectedError:
@@ -193,8 +217,8 @@ class LocalBox(BaseBox):
         return CodeBoxStatus(
             status="running"
             if self.kernel_id
-            and self.session
-            and (await self.session.get(self.kernel_url)).status == 200
+            and self.aiohttp_session
+            and (await self.aiohttp_session.get(self.kernel_url)).status == 200
             else "stopped"
         )
 
@@ -444,12 +468,14 @@ class LocalBox(BaseBox):
 
     def install(self, package_name: str) -> CodeBoxStatus:
         self.run(f"!pip install -q {package_name}")
-        # restart kernel if needed TODO
+        self.restart()
+        self.run(f"try:\n    import {package_name}\nexcept:\n    pass")
         return CodeBoxStatus(status=f"{package_name} installed successfully")
 
     async def ainstall(self, package_name: str) -> CodeBoxStatus:
         await self.arun(f"!pip install -q {package_name}")
-        # restart kernel if needed TODO
+        await self.arestart()
+        await self.arun(f"try:\n    import {package_name}\nexcept:\n    pass")
         return CodeBoxStatus(status=f"{package_name} installed successfully")
 
     def list_files(self) -> List[CodeBoxFile]:
@@ -476,39 +502,44 @@ class LocalBox(BaseBox):
         return CodeBoxStatus(status="restarted")
 
     def stop(self) -> CodeBoxStatus:
-        if self.ws is not None:
-            try:
-                self.ws.close()
-            except ConnectionClosedError:
-                pass
-            self.ws = None
-
         if self.jupyter is not None:
             self.jupyter.terminate()
             self.jupyter.wait()
             self.jupyter = None
             time.sleep(2)
 
-        return CodeBoxStatus(status="stopped")
-
-    async def astop(self) -> CodeBoxStatus:
         if self.ws is not None:
             try:
-                if not isinstance(self.ws, WebSocketClientProtocol):
-                    raise RuntimeError("Mixing asyncio and sync code is not supported")
-                await self.ws.close()
+                if isinstance(self.ws, ClientConnection):
+                    self.ws.close()
+                else:
+                    loop = asyncio.new_event_loop()
+                    loop.run_until_complete(self.ws.close())
             except ConnectionClosedError:
                 pass
             self.ws = None
 
+        return CodeBoxStatus(status="stopped")
+
+    async def astop(self) -> CodeBoxStatus:
         if self.jupyter is not None:
             self.jupyter.terminate()
             self.jupyter = None
             await asyncio.sleep(2)
 
-        if self.session is not None:
-            await self.session.close()
-            self.session = None
+        if self.ws is not None:
+            try:
+                if isinstance(self.ws, WebSocketClientProtocol):
+                    await self.ws.close()
+                else:
+                    self.ws.close()
+            except ConnectionClosedError:
+                pass
+            self.ws = None
+
+        if self.aiohttp_session is not None:
+            await self.aiohttp_session.close()
+            self.aiohttp_session = None
 
         return CodeBoxStatus(status="stopped")
 
@@ -521,3 +552,11 @@ class LocalBox(BaseBox):
     def ws_url(self) -> str:
         """Return the url of the websocket."""
         return f"ws://localhost:{self.port}/api"
+
+    def __del__(self):
+        self.stop()
+
+        if self.aiohttp_session is not None:
+            loop = asyncio.new_event_loop()
+            loop.run_until_complete(self.aiohttp_session.close())
+            self.aiohttp_session = None
